/** 
 * @file gl860-sysfs.c
 * @author Olivier LORIN, tiré du pilote Syntek par Nicolas VIVIEN
 * @date 2008-05-05
 * @version v0.9.0
 *
 * @brief Driver for GL860 USB video camera
 *
 * @note Copyright (C) Olivier LORIN
 *
 * @par Licences
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/errno.h>
#include <linux/slab.h>
#include <linux/kref.h>
#include <linux/device.h>

#include <linux/usb.h>
#include <media/v4l2-dev.h>

#include "gl860.h"


extern const struct gl860_coord wcam_image_sizes[WCAM_NBR_SIZES];


/** 
 * @brief show_release
 *
 * @param class Class device
 *
 * @retval buf Adress of buffer with the 'release' value
 * 
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t show_release(struct class_device *class, char *buf)
#else
static ssize_t show_release(struct device *class, struct device_attribute *attr, char *buf)
#endif
{
	struct video_device *vdev = to_video_device(class);
	struct usb_gl860 *dev = video_get_drvdata(vdev);

	return sprintf(buf, "%d\n", dev->release);
}


/** 
 * @brief show_videostatus
 *
 * @param class Class device
 *
 * @retval buf Adress of buffer with the 'videostatus' value
 * 
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t show_videostatus(struct class_device *class, char *buf)
#else
static ssize_t show_videostatus(struct device *class, struct device_attribute *attr, char *buf)
#endif
{
	struct video_device *vdev = to_video_device(class);
	struct usb_gl860 *dev = video_get_drvdata(vdev);

	return sprintf(buf,
			"Nbr ISOC errors    : %d\n"
			"Nbr dropped frames : %d\n"
			"Nbr dumped frames  : %d\n",
			dev->visoc_errors,
			dev->vframes_error,
			dev->vframes_dumped);
}


/** 
 * @brief show_informations
 *
 * @param class Class device
 *
 * @retval buf Adress of buffer with the 'informations' value
 * 
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t show_informations(struct class_device *class, char *buf)
#else
static ssize_t show_informations(struct device *class, struct device_attribute *attr, char *buf)
#endif
{
	int width, height;
	char *pixelfmt = NULL;

	struct video_device *vdev = to_video_device(class);
	struct usb_gl860 *dev = video_get_drvdata(vdev);

	char *palette_rgb24 = "RGB24 - RGB-8-8-8 - 24 bits";
	char *palette_rgb32 = "RGB32 - RGB-8-8-8-8 - 32 bits";
	char *palette_bgr24 = "BGR24 - BGR-8-8-8 - 24 bits";
	char *palette_bgr32 = "BGR32 - BGR-8-8-8-8 - 32 bits";
	char *palette_uyvy = "UYVY - YUV 4:2:2 - 16 bits";
	char *palette_yuyv = "YUYV - YUV 4:2:2 - 16 bits";


	switch (dev->vsettings.palette) {
		case WCAM_PALETTE_RGB24:
			pixelfmt = palette_rgb24;
			break;

		case WCAM_PALETTE_RGB32:
			pixelfmt = palette_rgb32;
			break;

		case WCAM_PALETTE_BGR24:
			pixelfmt = palette_bgr24;
			break;

		case WCAM_PALETTE_BGR32:
			pixelfmt = palette_bgr32;
			break;

		case WCAM_PALETTE_UYVY:
			pixelfmt = palette_uyvy;
			break;

		case WCAM_PALETTE_YUYV:
			pixelfmt = palette_yuyv;
			break;
	}

	switch (dev->resolution) {
		case WCAM_80x60:
		case WCAM_128x96:
		case WCAM_160x120:
		case WCAM_213x160:
		case WCAM_320x240:
		case WCAM_640x480:
			width = wcam_image_sizes[WCAM_640x480].x;
			height = wcam_image_sizes[WCAM_640x480].y;
			break;

		case WCAM_800x600:
			width = wcam_image_sizes[WCAM_800x600].x;
			height = wcam_image_sizes[WCAM_800x600].y;
			break;

//		case WCAM_1024x768:
		case WCAM_1280x960:
			width = wcam_image_sizes[WCAM_1280x960].x;
			height = wcam_image_sizes[WCAM_1280x960].y;
			break;

		default:
			width = 0;
			height = 0;
	}

	return sprintf(buf,
			"Asked resolution  : %dx%d\n"
			"Driver resolution : %dx%d\n"
			"Webcam resolution : %dx%d\n"
			"\n"
			"%s\n"
			"\n"
			"Lum               : 0x%X\n"
			"Src Lum           : 0x%X\n"
			"Couleur           : 0x%X\n"
			"Nettete           : 0x%X\n"
			"Contrast          : 0x%X\n"
			"Contre-jour       : 0x%X\n",
			dev->view.x, dev->view.y,
			wcam_image_sizes[dev->resolution].x, wcam_image_sizes[dev->resolution].y,
			width, height,
			pixelfmt,
			0xFFFF & dev->vsettings.brightness,
			0xFFFF & dev->vsettings.srclum,
			0xFFFF & dev->vsettings.colour,
			0xFFFF & dev->vsettings.net,
			0xFFFF & dev->vsettings.contrast,
			0xFFFF & dev->vsettings.contrej);
}


/** 
 * @brief show_fps
 *
 * @param class Class device
 *
 * @retval buf Adress of buffer with the 'fps' value
 * 
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t show_fps(struct class_device *class, char *buf)
#else
static ssize_t show_fps(struct device *class, struct device_attribute *attr, char *buf)
#endif
{
	struct video_device *vdev = to_video_device(class);
	struct usb_gl860 *dev = video_get_drvdata(vdev);

	return sprintf(buf, "%d\n", dev->vsettings.fps);
}


/** 
 * @brief show_brightness
 *
 * @param class Class device
 *
 * @retval buf Adress of buffer with the 'brightness' value
 * 
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t show_brightness(struct class_device *class, char *buf)
#else
static ssize_t show_brightness(struct device *class, struct device_attribute *attr, char *buf)
#endif
{
	struct video_device *vdev = to_video_device(class);
	struct usb_gl860 *dev = video_get_drvdata(vdev);

	return sprintf(buf, "%X\n", dev->vsettings.brightness);
}


/** 
 * @brief store_brightness
 *
 * @param class Class device
 * @param buf Buffer
 * @param count Counter
 *
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t store_brightness(struct class_device *class, const char *buf, size_t count)
#else
static ssize_t store_brightness(struct device *class, struct device_attribute *attr, 
		const char *buf, size_t count)
#endif
{
	char *endp;
	unsigned long value;

	struct video_device *vdev = to_video_device(class);
	struct usb_gl860 *dev = video_get_drvdata(vdev);

	value = simple_strtoul(buf, &endp, 16);

	dev->vsettings.brightness = (int) value;

	dev_gl860_camera_settings(dev);

	return strlen(buf);
}

/** 
 * @brief show_contrast
 *
 * @param class Class device
 *
 * @retval buf Adress of buffer with the 'contrast' value
 * 
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t show_contrast(struct class_device *class, char *buf)
#else
static ssize_t show_contrast(struct device *class, struct device_attribute *attr, char *buf)
#endif
{
	struct video_device *vdev = to_video_device(class);
	struct usb_gl860 *dev = video_get_drvdata(vdev);

	return sprintf(buf, "%X\n", dev->vsettings.contrast);
}


/** 
 * @brief store_contrast
 *
 * @param class Class device
 * @param buf Buffer
 * @param count Counter
 *
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t store_contrast(struct class_device *class, const char *buf, size_t count)
#else
static ssize_t store_contrast(struct device *class, struct device_attribute *attr,
		const char *buf, size_t count)
#endif
{
	char *endp;
	unsigned long value;

	struct video_device *vdev = to_video_device(class);
	struct usb_gl860 *dev = video_get_drvdata(vdev);

	value = simple_strtoul(buf, &endp, 16);

	dev->vsettings.contrast = (int) value;

	dev_gl860_set_camera_quality(dev);

	return strlen(buf);
}


/** 
 * @brief show_whitebalance
 *
 * @param class Class device
 *
 * @retval buf Adress of buffer with the 'whitebalance' value
 * 
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t show_whitebalance(struct class_device *class, char *buf)
#else
static ssize_t show_whitebalance(struct device *class, struct device_attribute *attr, char *buf)
#endif
{
	struct video_device *vdev = to_video_device(class);
	struct usb_gl860 *dev = video_get_drvdata(vdev);

	return sprintf(buf, "%X\n", dev->vsettings.whiteness);
}


/** 
 * @brief store_whitebalance
 *
 * @param class Class device
 * @param buf Buffer
 * @param count Counter
 *
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t store_whitebalance(struct class_device *class, const char *buf, size_t count)
#else
static ssize_t store_whitebalance(struct device *class, struct device_attribute *attr,
		const char *buf, size_t count)
#endif
{
	char *endp;
	unsigned long value;

	struct video_device *vdev = to_video_device(class);
	struct usb_gl860 *dev = video_get_drvdata(vdev);

	value = simple_strtoul(buf, &endp, 16);

	dev->vsettings.whiteness = (int) value;

	dev_gl860_set_camera_quality(dev);

	return strlen(buf);
}


/** 
 * @brief show_colour
 *
 * @param class Class device
 *
 * @retval buf Adress of buffer with the 'colour' value
 * 
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t show_colour(struct class_device *class, char *buf)
#else
static ssize_t show_colour(struct device *class, struct device_attribute *attr, char *buf)
#endif
{
	struct video_device *vdev = to_video_device(class);
	struct usb_gl860 *dev = video_get_drvdata(vdev);

	return sprintf(buf, "%X\n", dev->vsettings.colour);
}


/** 
 * @brief store_colour
 *
 * @param class Class device
 * @param buf Buffer
 * @param count Counter
 *
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t store_colour(struct class_device *class, const char *buf, size_t count)
#else
static ssize_t store_colour(struct device *class, struct device_attribute *attr,
		const char *buf, size_t count)
#endif
{
	char *endp;
	unsigned long value;

	struct video_device *vdev = to_video_device(class);
	struct usb_gl860 *dev = video_get_drvdata(vdev);

	value = simple_strtoul(buf, &endp, 16);

	dev->vsettings.colour = (int) value;

	dev_gl860_camera_settings(dev);

	return strlen(buf);
}


/** 
 * @brief show_hflip
 *
 * @param class Class device
 *
 * @retval buf Adress of buffer with the 'hflip' value
 * 
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t show_hflip(struct class_device *class, char *buf)
#else
static ssize_t show_hflip(struct device *class, struct device_attribute *attr, char *buf)
#endif
{
	struct video_device *vdev = to_video_device(class);
	struct usb_gl860 *dev = video_get_drvdata(vdev);

	return sprintf(buf, "%d\n", dev->vsettings.miroirh);
}


/** 
 * @brief store_hflip
 *
 * @param class Class device
 * @param buf Buffer
 * @param count Counter
 *
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t store_hflip(struct class_device *class, const char *buf, size_t count)
#else
static ssize_t store_hflip(struct device *class, struct device_attribute *attr,
		const char *buf, size_t count)
#endif
{
	struct video_device *vdev = to_video_device(class);
	struct usb_gl860 *dev = video_get_drvdata(vdev);

	if (strncmp(buf, "1", 1) == 0)
		dev->vsettings.miroirh = 1;
	else if (strncmp(buf, "0", 1) == 0)
		dev->vsettings.miroirh = 0;
	else
		return -EINVAL;

	dev_gl860_camera_settings(dev);

	return strlen(buf);
}


/** 
 * @brief show_vflip
 *
 * @param class Class device
 *
 * @retval buf Adress of buffer with the 'vflip' value
 * 
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t show_vflip(struct class_device *class, char *buf)
#else
static ssize_t show_vflip(struct device *class, struct device_attribute *attr, char *buf)
#endif
{
	struct video_device *vdev = to_video_device(class);
	struct usb_gl860 *dev = video_get_drvdata(vdev);

	return sprintf(buf, "%d\n", dev->vsettings.miroirv);
}


/** 
 * @brief store_vflip
 *
 * @param class Class device
 * @param buf Buffer
 * @param count Counter
 *
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t store_vflip(struct class_device *class, const char *buf, size_t count)
#else
static ssize_t store_vflip(struct device *class, struct device_attribute *attr, const char *buf, size_t count)
#endif
{
	struct video_device *vdev = to_video_device(class);
	struct usb_gl860 *dev = video_get_drvdata(vdev);

	if (strncmp(buf, "1", 1) == 0)
		dev->vsettings.miroirv = 1;
	else if (strncmp(buf, "0", 1) == 0)
		dev->vsettings.miroirv = 0;
	else
		return -EINVAL;

	dev_gl860_camera_settings(dev);

	return strlen(buf);
}



/** 
 * @brief show_saturation
 *
 * @param class Class device
 *
 * @retval buf Adress of buffer with the 'saturation' value
 * 
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t show_saturation(struct class_device *class, char *buf)
#else
static ssize_t show_saturation(struct device *class, struct device_attribute *attr, char *buf)
#endif
{
	struct video_device *vdev = to_video_device(class);
	struct usb_gl860 *dev = video_get_drvdata(vdev);

	return sprintf(buf, "%X\n", dev->vsettings.saturation);
}


/** 
 * @brief store_saturation
 *
 * @param class Class device
 * @param buf Buffer
 * @param count Counter
 *
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t store_saturation(struct class_device *class, const char *buf, size_t count)
#else
static ssize_t store_saturation(struct device *class, struct device_attribute *attr,
		const char *buf, size_t count)
#endif
{
	char *endp;
	unsigned long value;

	struct video_device *vdev = to_video_device(class);
	struct usb_gl860 *dev = video_get_drvdata(vdev);

	value = simple_strtoul(buf, &endp, 16);

	dev->vsettings.saturation = (int) value;

	dev_gl860_camera_settings(dev);

	return strlen(buf);
}



/** 
 * @brief show_srclum
 *
 * @param class Class device
 *
 * @retval buf Adress of buffer with the 'srclum' value
 * 
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t show_srclum(struct class_device *class, char *buf)
#else
static ssize_t show_srclum(struct device *class, struct device_attribute *attr, char *buf)
#endif
{
	struct video_device *vdev = to_video_device(class);
	struct usb_gl860 *dev = video_get_drvdata(vdev);

	return sprintf(buf, "%X\n", dev->vsettings.srclum);
}


/** 
 * @brief store_srclum
 *
 * @param class Class device
 * @param buf Buffer
 * @param count Counter
 *
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t store_srclum(struct class_device *class, const char *buf, size_t count)
#else
static ssize_t store_srclum(struct device *class, struct device_attribute *attr,
		const char *buf, size_t count)
#endif
{
	char *endp;
	unsigned long value;

	struct video_device *vdev = to_video_device(class);
	struct usb_gl860 *dev = video_get_drvdata(vdev);

	value = simple_strtoul(buf, &endp, 16);

	dev->vsettings.srclum = (int) value;

	dev_gl860_camera_settings(dev);

	return strlen(buf);
}



/** 
 * @brief show_freq50
 *
 * @param class Class device
 *
 * @retval buf Adress of buffer with the 'freq50' value
 * 
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t show_freq50(struct class_device *class, char *buf)
#else
static ssize_t show_freq50(struct device *class, struct device_attribute *attr, char *buf)
#endif
{
   struct video_device *vdev = to_video_device(class);
   struct usb_gl860 *dev = video_get_drvdata(vdev);

   return sprintf(buf, "%d\n", dev->vsettings.freq50);
}

/** 
 * @brief store_freq50
 *
 * @param class Class device
 * @param buf Buffer
 * @param count Counter
 *
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t store_freq50(struct class_device *class, const char *buf, size_t count)
#else
static ssize_t store_freq50(struct device *class, struct device_attribute *attr, const char *buf, size_t count)
#endif
{
	struct video_device *vdev = to_video_device(class);
	struct usb_gl860 *dev = video_get_drvdata(vdev);

	if (strncmp(buf, "1", 1) == 0)
		dev->vsettings.freq50 = 1;
	else if (strncmp(buf, "0", 1) == 0)
		dev->vsettings.freq50 = 0;
	else
		return -EINVAL;

	dev_gl860_camera_settings(dev);

	return strlen(buf);
}



/** 
 * @brief show_nettete
 *
 * @param class Class device
 *
 * @retval buf Adress of buffer with the 'nettete' value
 * 
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t show_nettete(struct class_device *class, char *buf)
#else
static ssize_t show_nettete(struct device *class, struct device_attribute *attr, char *buf)
#endif
{
	struct video_device *vdev = to_video_device(class);
	struct usb_gl860 *dev = video_get_drvdata(vdev);

	return sprintf(buf, "%X\n", dev->vsettings.net);
}


/** 
 * @brief store_nettete
 *
 * @param class Class device
 * @param buf Buffer
 * @param count Counter
 *
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t store_nettete(struct class_device *class, const char *buf, size_t count)
#else
static ssize_t store_nettete(struct device *class, struct device_attribute *attr,
		const char *buf, size_t count)
#endif
{
	char *endp;
	unsigned long value;

	struct video_device *vdev = to_video_device(class);
	struct usb_gl860 *dev = video_get_drvdata(vdev);

	value = simple_strtoul(buf, &endp, 16);

	dev->vsettings.net = (int) value;

	dev_gl860_camera_settings(dev);

	return strlen(buf);
}


/** 
 * @brief show_contrej
 *
 * @param class Class device
 *
 * @retval buf Adress of buffer with the 'contrej' value
 * 
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t show_contrej(struct class_device *class, char *buf)
#else
static ssize_t show_contrej(struct device *class, struct device_attribute *attr, char *buf)
#endif
{
	struct video_device *vdev = to_video_device(class);
	struct usb_gl860 *dev = video_get_drvdata(vdev);

	return sprintf(buf, "%X\n", dev->vsettings.contrej);
}


/** 
 * @brief store_contrej
 *
 * @param class Class device
 * @param buf Buffer
 * @param count Counter
 *
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t store_contrej(struct class_device *class, const char *buf, size_t count)
#else
static ssize_t store_contrej(struct device *class, struct device_attribute *attr,
		const char *buf, size_t count)
#endif
{
	char *endp;
	unsigned long value;

	struct video_device *vdev = to_video_device(class);
	struct usb_gl860 *dev = video_get_drvdata(vdev);

	value = simple_strtoul(buf, &endp, 16);

	dev->vsettings.contrej = (int) value;

	dev_gl860_camera_settings(dev);

	return strlen(buf);
}



/** 
 * @brief show_gamma
 *
 * @param class Class device
 *
 * @retval buf Adress of buffer with the 'gamma' value
 * 
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t show_gamma(struct class_device *class, char *buf)
#else
static ssize_t show_gamma(struct device *class, struct device_attribute *attr, char *buf)
#endif
{
	struct video_device *vdev = to_video_device(class);
	struct usb_gl860 *dev = video_get_drvdata(vdev);

	return sprintf(buf, "%X\n", dev->vsettings.whiteness);
}


/** 
 * @brief store_gamma
 *
 * @param class Class device
 * @param buf Buffer
 * @param count Counter
 *
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t store_gamma(struct class_device *class, const char *buf, size_t count)
#else
static ssize_t store_gamma(struct device *class, struct device_attribute *attr, 
		const char *buf, size_t count)
#endif
{
	char *endp;
	unsigned long value;

	struct video_device *vdev = to_video_device(class);
	struct usb_gl860 *dev = video_get_drvdata(vdev);

	value = simple_strtoul(buf, &endp, 16);

	dev->vsettings.whiteness = (int) value;

	dev_gl860_camera_settings(dev);

	return strlen(buf);
}





#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static CLASS_DEVICE_ATTR(release, S_IRUGO, show_release, NULL);										/**< Release value */
static CLASS_DEVICE_ATTR(videostatus, S_IRUGO, show_videostatus, NULL);								/**< Video status */
static CLASS_DEVICE_ATTR(informations, S_IRUGO, show_informations, NULL);							/**< Informations */
static CLASS_DEVICE_ATTR(fps, S_IRUGO, show_fps, NULL);												/**< FPS value */
static CLASS_DEVICE_ATTR(brightness, S_IRUGO | S_IWUGO, show_brightness, store_brightness);			/**< Brightness value */
static CLASS_DEVICE_ATTR(contrast, S_IRUGO | S_IWUGO, show_contrast, store_contrast);				/**< Contrast value */
static CLASS_DEVICE_ATTR(whitebalance, S_IRUGO | S_IWUGO, show_whitebalance, store_whitebalance);	/**< Whitebalance value */
static CLASS_DEVICE_ATTR(colour, S_IRUGO | S_IWUGO, show_colour, store_colour);						/**< Hue value */
static CLASS_DEVICE_ATTR(hflip, S_IRUGO | S_IWUGO, show_hflip, store_hflip);						/**< Horizontal filp value */
static CLASS_DEVICE_ATTR(vflip, S_IRUGO | S_IWUGO, show_vflip, store_vflip);						/**< Vertical filp value */

static CLASS_DEVICE_ATTR(saturation, S_IRUGO | S_IWUGO, show_saturation, store_saturation);		/**< saturation value */
static CLASS_DEVICE_ATTR(srclum, S_IRUGO | S_IWUGO, show_srclum, store_srclum);						/**< src de lum.  value */
static CLASS_DEVICE_ATTR(freq50, S_IRUGO | S_IWUGO, show_freq50, store_freq50);						/**< freq50  value */
static CLASS_DEVICE_ATTR(nettete, S_IRUGO | S_IWUGO, show_nettete, store_nettete);					/**< netteté value */
static CLASS_DEVICE_ATTR(contrej, S_IRUGO | S_IWUGO, show_contrej, store_contrej);					/**< contre-jour value */
static CLASS_DEVICE_ATTR(gamma, S_IRUGO | S_IWUGO, show_gamma, store_gamma);					/**< contre-jour value */
#else
static DEVICE_ATTR(release, S_IRUGO, show_release, NULL);											/**< Release value */
static DEVICE_ATTR(videostatus, S_IRUGO, show_videostatus, NULL);									/**< Video status */
static DEVICE_ATTR(informations, S_IRUGO, show_informations, NULL);									/**< Informations */
static DEVICE_ATTR(fps, S_IRUGO, show_fps, NULL);													/**< FPS value */
static DEVICE_ATTR(brightness, S_IRUGO | S_IWUGO, show_brightness, store_brightness);				/**< Brightness value */
static DEVICE_ATTR(contrast, S_IRUGO | S_IWUGO, show_contrast, store_contrast);						/**< Contrast value */
static DEVICE_ATTR(whitebalance, S_IRUGO | S_IWUGO, show_whitebalance, store_whitebalance);			/**< Whitebalance value */
static DEVICE_ATTR(colour, S_IRUGO | S_IWUGO, show_colour, store_colour);							/**< Hue value */
static DEVICE_ATTR(hflip, S_IRUGO | S_IWUGO, show_hflip, store_hflip);								/**< Horizontal filp value */
static DEVICE_ATTR(vflip, S_IRUGO | S_IWUGO, show_vflip, store_vflip);								/**< Vertical filp value */

static DEVICE_ATTR(saturation, S_IRUGO | S_IWUGO, show_saturation, store_saturation);		/**< saturation value */
static DEVICE_ATTR(srclum, S_IRUGO | S_IWUGO, show_srclum, store_srclum);						/**< src de lum.  value */
static DEVICE_ATTR(freq50, S_IRUGO | S_IWUGO, show_freq50, store_freq50);						/**< freq50  value */
static DEVICE_ATTR(nettete, S_IRUGO | S_IWUGO, show_nettete, store_nettete);					/**< netteté value */
static DEVICE_ATTR(contrej, S_IRUGO | S_IWUGO, show_contrej, store_contrej);					/**< contre-jour value */
static DEVICE_ATTR(gamma, S_IRUGO | S_IWUGO, show_gamma, store_gamma);					/**< contre-jour value */
#endif


/** 
 * @brief Create the 'sys' entries.
 *
 * This function permits to create all the entries in the 'sys' filesystem.
 *
 * @param vdev Video device structure
 * 
 * @returns 0 if all is OK
 */
int gl860_create_sysfs_files(struct video_device *vdev)
{
	int ret;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
	ret = video_device_create_file(vdev, &class_device_attr_release);
	ret = video_device_create_file(vdev, &class_device_attr_videostatus);
	ret = video_device_create_file(vdev, &class_device_attr_informations);
	ret = video_device_create_file(vdev, &class_device_attr_fps);
	ret = video_device_create_file(vdev, &class_device_attr_brightness);
	ret = video_device_create_file(vdev, &class_device_attr_contrast);
	ret = video_device_create_file(vdev, &class_device_attr_whitebalance);
	ret = video_device_create_file(vdev, &class_device_attr_colour);
	ret = video_device_create_file(vdev, &class_device_attr_hflip);
	ret = video_device_create_file(vdev, &class_device_attr_vflip);

	ret = video_device_create_file(vdev, &class_device_attr_saturation);
	ret = video_device_create_file(vdev, &class_device_attr_srclum);
	ret = video_device_create_file(vdev, &class_device_attr_freq50);
	ret = video_device_create_file(vdev, &class_device_attr_nettete);
	ret = video_device_create_file(vdev, &class_device_attr_contrej);
	ret = video_device_create_file(vdev, &class_device_attr_gamma);
#else
	ret = device_create_file(&vdev->dev, &dev_attr_release);
	ret = device_create_file(&vdev->dev, &dev_attr_videostatus);
	ret = device_create_file(&vdev->dev, &dev_attr_informations);
	ret = device_create_file(&vdev->dev, &dev_attr_fps);
	ret = device_create_file(&vdev->dev, &dev_attr_brightness);
	ret = device_create_file(&vdev->dev, &dev_attr_contrast);
	ret = device_create_file(&vdev->dev, &dev_attr_whitebalance);
	ret = device_create_file(&vdev->dev, &dev_attr_colour);
	ret = device_create_file(&vdev->dev, &dev_attr_hflip);
	ret = device_create_file(&vdev->dev, &dev_attr_vflip);

	ret = device_create_file(&vdev->dev, &dev_attr_saturation);
	ret = device_create_file(&vdev->dev, &dev_attr_srclum);
	ret = device_create_file(&vdev->dev, &dev_attr_freq50);
	ret = device_create_file(&vdev->dev, &dev_attr_nettete);
	ret = device_create_file(&vdev->dev, &dev_attr_contrej);
	ret = device_create_file(&vdev->dev, &dev_attr_gamma);
#endif

	return ret;
}


/** 
 * @brief Remove the 'sys' entries.
 *
 * This function permits to remove all the entries in the 'sys' filesystem.
 *
 * @param vdev Video device structure
 * 
 * @returns 0 if all is OK
 */
void gl860_remove_sysfs_files(struct video_device *vdev)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
	video_device_remove_file(vdev, &class_device_attr_release);
	video_device_remove_file(vdev, &class_device_attr_videostatus);
	video_device_remove_file(vdev, &class_device_attr_informations);
	video_device_remove_file(vdev, &class_device_attr_fps);
	video_device_remove_file(vdev, &class_device_attr_brightness);
	video_device_remove_file(vdev, &class_device_attr_contrast);
	video_device_remove_file(vdev, &class_device_attr_whitebalance);
	video_device_remove_file(vdev, &class_device_attr_colour);
	video_device_remove_file(vdev, &class_device_attr_hflip);
	video_device_remove_file(vdev, &class_device_attr_vflip);

	video_device_remove_file(vdev, &class_device_attr_saturation);
	video_device_remove_file(vdev, &class_device_attr_srclum);
	video_device_remove_file(vdev, &class_device_attr_freq50);
	video_device_remove_file(vdev, &class_device_attr_nettete);
	video_device_remove_file(vdev, &class_device_attr_contrej);
	video_device_remove_file(vdev, &class_device_attr_gamma);
#else
	device_remove_file(&vdev->dev, &dev_attr_release);
	device_remove_file(&vdev->dev, &dev_attr_videostatus);
	device_remove_file(&vdev->dev, &dev_attr_informations);
	device_remove_file(&vdev->dev, &dev_attr_fps);
	device_remove_file(&vdev->dev, &dev_attr_brightness);
	device_remove_file(&vdev->dev, &dev_attr_contrast);
	device_remove_file(&vdev->dev, &dev_attr_whitebalance);
	device_remove_file(&vdev->dev, &dev_attr_colour);
	device_remove_file(&vdev->dev, &dev_attr_hflip);
	device_remove_file(&vdev->dev, &dev_attr_vflip);

	device_remove_file(&vdev->dev, &dev_attr_saturation);
	device_remove_file(&vdev->dev, &dev_attr_srclum);
	device_remove_file(&vdev->dev, &dev_attr_freq50);
	device_remove_file(&vdev->dev, &dev_attr_nettete);
	device_remove_file(&vdev->dev, &dev_attr_contrej);
	device_remove_file(&vdev->dev, &dev_attr_gamma);
#endif
}

