/** 
 * @file gl860-v4l.c
 * @author Olivier LORIN, tiré du pilote Syntek par Nicolas VIVIEN
 * @date 2008-05-05
 * @version v0.9.0
 *
 * @brief Driver for GL860 USB video camera
 *
 * @note Copyright (C) Olivier LORIN
 *
 * @par Licences
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/errno.h>
#include <linux/slab.h>
#include <linux/kref.h>
#include <linux/vmalloc.h>
#include <linux/pagemap.h>

#include <linux/usb.h>
#include <media/v4l2-common.h>
#include <media/v4l2-dev.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24)
#include <media/v4l2-ioctl.h>
#endif
//#include <linux/videodev2.h>
#include <linux/videodev.h>

#include "gl860.h"

#define V4L2_CID_SRC_LUM (V4L2_CID_PRIVATE_BASE + 0)
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
	#define V4L2_CID_POWER_LINE_FREQUENCY (V4L2_CID_PRIVATE_BASE + 1)
	#define V4L2_CID_SHARPNESS (V4L2_CID_PRIVATE_BASE + 2)
	#define V4L2_CID_BACKLIGHT_COMPENSATION (V4L2_CID_PRIVATE_BASE + 3)
#endif

static struct file_operations v4l_gl860_fops;


/**
 * @var wcam_image_sizes
 *   List of all resolutions supported by the driver
 */
struct gl860_coord wcam_image_sizes[WCAM_NBR_SIZES] = {
	{   80,   60 },
	{  128,   96 },
	{  160,  120 },
	{  213,  160 },
	{  320,  240 },
	{  640,  480 },
	{  800,  600 },
	{ 1280,  960 },
	{ 1600, 1200 }
};

static int nbVGA=0, nbSXGA=0, nbUXGA=0;

/**
 * @var gl860_controls
 *   List of all V4Lv2 controls supported by the driver
 */
static struct v4l2_queryctrl gl860_controls[] = {
//#if defined(_F191_) || defined(_0503A_) || defined(_0503B_) || defined(_0503C_)
	{
		.id      = V4L2_CID_HFLIP,
		.type    = V4L2_CTRL_TYPE_BOOLEAN,
		.name    = LBL_MIROIR_H,
		.minimum = 0,
		.maximum = 1,
		.step    = 1,
		.default_value = 0,
	},
	{
		.id      = V4L2_CID_VFLIP,
		.type    = V4L2_CTRL_TYPE_BOOLEAN,
		.name    = LBL_MIROIR_V,
		.minimum = 0,
		.maximum = 1,
		.step    = 1,
		.default_value = 0,
	},
//#endif
//#if defined (_F191_) || defined (_0503A_) || defined(_0503B_) || defined(_0503C_)
	{
		.id      = V4L2_CID_POWER_LINE_FREQUENCY,
		.type    = V4L2_CTRL_TYPE_BOOLEAN,
		.name    = "50 Hz",
		.minimum = 0,
		.maximum = 1,
		.step    = 1,
		.default_value = 1,
	},
//#endif
	{
		.id      = V4L2_CID_BRIGHTNESS,
		.type    = V4L2_CTRL_TYPE_INTEGER,
		.name    = LBL_LUMINOSITE,
		.minimum = 0,
		.maximum = 0xFF00,
		.step    = 1,
		.default_value = 0,
	},
//#ifndef _0503SIM_
	{
		.id      = V4L2_CID_SATURATION,
		.type    = V4L2_CTRL_TYPE_INTEGER,
		.name    = "Saturation",
		.minimum = 0,
		.maximum = 0xFF00,
		.step    = 1,
		.default_value = 0xFF00,
	},
	{
		.id      = V4L2_CID_HUE,
		.type    = V4L2_CTRL_TYPE_INTEGER,
		.name    = "Palette",
		.minimum = 0,
		.maximum = 0xFF00,
		.step    = 1,
		.default_value = 0,
	},
	{
		.id      = V4L2_CID_SHARPNESS,
		.type    = V4L2_CTRL_TYPE_INTEGER,
		.name    = LBL_NETTETE,
		.minimum = 0,
		.maximum = 0xFF00,
		.step    = 1,
		.default_value = 0xFF00,
	},
	{
		.id      = V4L2_CID_BACKLIGHT_COMPENSATION,
		.type    = V4L2_CTRL_TYPE_INTEGER,
		.name    = LBL_CONTREJ,
		.minimum = 0,
		.maximum = 2,
		.step    = 1,
		.default_value = 1,
	},
	{
		.id      = V4L2_CID_SRC_LUM,
		.type    = V4L2_CTRL_TYPE_INTEGER,
		.name    = LBL_SRC_LUM,
		.minimum = 0,
		.maximum = 2,
		.step    = 1,
		.default_value = 0,
	},
//#endif
//#ifndef _F191_
	{
		.id      = V4L2_CID_WHITENESS,
		.type    = V4L2_CTRL_TYPE_INTEGER,
		.name    = LBL_BLANCHEUR,
		.minimum = 0,
		.maximum = 0xff00,
		.step    = 1,
		.default_value = 0x7f00,
	},
//#endif
	{
		.id      = V4L2_CID_CONTRAST,
		.type    = V4L2_CTRL_TYPE_INTEGER,
		.name    = LBL_CONTRASTE,
		.minimum = 0,
		.maximum = 0xff00,
		.step    = 1,
		.default_value = 0x7f00,
	},
////#endif
};


/** 
 * @param dev
 * @param width Width of wished resolution
 * @param height Height of wished resolution
 * 
 * @returns 0 if all is OK
 *
 * @brief Select a video mode
 *
 * This function permits to check and select a video mode.
 */
int v4l_gl860_select_video_mode(struct usb_gl860 *dev, int width, int height)
{
	int i;
	int find;

	STK_DEV("Demande de la taille %dx%d\n",width,height);


	// Check width and height
	// Notice : this test is usefull for the Kopete application !

	// Driver can't build an image smaller than the minimal resolution !
	if ((width < wcam_image_sizes[0].x)
			|| (height < wcam_image_sizes[0].y)) {
		width = wcam_image_sizes[0].x;
		height = wcam_image_sizes[0].y;
	}


	// Driver can't build an image bigger than the maximal resolution !
	switch (dev->webcam_type) {
		case WCAM_VGA:
			if ((   width > wcam_image_sizes[nbVGA].x)
				|| (height > wcam_image_sizes[nbVGA].y)) {
				width  = wcam_image_sizes[nbVGA].x;
				height = wcam_image_sizes[nbVGA].y;
			}
			break;

		case WCAM_SXGA:
			if ((   width > wcam_image_sizes[nbSXGA].x)
				|| (height > wcam_image_sizes[nbSXGA].y)) {
				width  = wcam_image_sizes[nbSXGA].x;
				height = wcam_image_sizes[nbSXGA].y;
			}
			break;

		case WCAM_UXGA:
			if ((   width > wcam_image_sizes[nbUXGA].x)
				|| (height > wcam_image_sizes[nbUXGA].y)) {
				width  = wcam_image_sizes[nbUXGA].x;
				height = wcam_image_sizes[nbUXGA].y;
			}
			break;

		default:
			return -1;
	}


	// Seek the best resolution
	switch (dev->webcam_type) {
		case WCAM_UXGA:
			for (i=0, find=0; i<=nbUXGA; i++) {
				if (wcam_image_sizes[i].x < width && wcam_image_sizes[i].y < height)
					find = i+1;
			}
			break;

		case WCAM_SXGA:
			for (i=0, find=0; i<=nbSXGA; i++) {
				if (wcam_image_sizes[i].x < width && wcam_image_sizes[i].y < height)
					find = i+1;
			}
			break;

		case WCAM_VGA:
			for (i=0, find=0; i<=nbVGA; i++) {
				if (wcam_image_sizes[i].x < width && wcam_image_sizes[i].y < height)
					find = i+1;
			}
			break;

		default:
			return -1;
	}

	// Save the new resolution
	dev->resolution = find;

	STK_DEBUG("Set mode %d [%dx%d]\n", dev->resolution,
			wcam_image_sizes[dev->resolution].x, wcam_image_sizes[dev->resolution].y);

	// Save the new size
	dev->view.x = width;
	dev->view.y = height;


	// Calculate the frame size
	switch (dev->resolution) {
		case WCAM_80x60:
		case WCAM_128x96:
		case WCAM_160x120:
		case WCAM_213x160:
		case WCAM_320x240:
		case WCAM_640x480:
			dev->image.x = wcam_image_sizes[WCAM_640x480].x;
			dev->image.y = wcam_image_sizes[WCAM_640x480].y;
			dev->frame_size = dev->image.x * dev->image.y;
			break;

		case WCAM_800x600:
			dev->image.x = wcam_image_sizes[WCAM_800x600].x;
			dev->image.y = wcam_image_sizes[WCAM_800x600].y;
			dev->frame_size = dev->image.x * dev->image.y;
			break;

		case WCAM_1280x960:
			dev->image.x = wcam_image_sizes[WCAM_1280x960].x;
			dev->image.y = wcam_image_sizes[WCAM_1280x960].y;
			dev->frame_size = dev->image.x * dev->image.y;
			break;

		case WCAM_1600x1200:
			dev->image.x = wcam_image_sizes[WCAM_1600x1200].x;
			dev->image.y = wcam_image_sizes[WCAM_1600x1200].y;
			dev->frame_size = dev->image.x * dev->image.y;
			break;
	}


	// Calculate the image size
	switch (dev->vsettings.palette) {
		case WCAM_PALETTE_RGB24:
		case WCAM_PALETTE_BGR24:
			//dev->view_size = ((3 * dev->view.x -1)/4 +1)*4 * dev->view.y;
			dev->view_size = 3 * dev->view.x * dev->view.y;
			dev->image_size = 3 * dev->frame_size;
			break;

		case WCAM_PALETTE_GRIS:
			dev->view_size = dev->view.x * dev->view.y;
			dev->image_size = dev->frame_size;
			break;

		case WCAM_PALETTE_RGB32:
		case WCAM_PALETTE_BGR32:
			dev->view_size = 3 * dev->view.x * dev->view.y;
			dev->image_size = 4 * dev->frame_size;
			break;

		case WCAM_PALETTE_UYVY:
		case WCAM_PALETTE_YUYV:
			dev->view_size = 2 * dev->view.x * dev->view.y;
			dev->image_size = 2 * dev->frame_size;
			break;
	}

	dev->framebuf->filled=0;

	return 0;
}


/** 
 * @param inode Pointer on an inode
 * @param fp File pointer
 * 
 * @returns 0 if all is OK
 *
 * @brief Open the video device
 *
 * This function permits to open a video device (/dev/videoX)
 */
static int v4l_gl860_open(struct inode *inode, struct file *fp)
{
	int err, i;

	struct usb_gl860 *dev;
	struct video_device *vdev;
	
	vdev = video_devdata(fp);
	dev = video_get_drvdata(video_devdata(fp));

	if (dev == NULL) {
		STK_ERROR("Device not initialized !!!\n");
		BUG();
	}

	if (dev->vopen) {
		STK_DEBUG("Device is busy, someone is using the device\n");
		return -EBUSY;
	}

	// Allocate memory
	err = gl860_allocate_buffers(dev);

	if (err < 0) {
		STK_ERROR("Failed to allocate buffer memory !\n");
		return err;
	}
	
	// Reset buffers and parameters
	gl860_reset_buffers(dev);

	// Settings
	dev->error_status = 0;
	dev->visoc_errors = 0;
	dev->vframes_error = 0;
	dev->vframes_dumped = 0;
	dev->vsettings.hue = 0;
	dev->vsettings.depth = 24;
	dev->vsettings.palette = WCAM_PALETTE_RGB24; //État initial de l'image de la caméra //AZERTY BGR

	dev->en_fct=0;
/*
	// Select the resolution by default
	v4l_gl860_select_video_mode(dev, 640, 480);


	// Initialize the device
	dev_gl860_init_camera(dev);
	dev_gl860_reconf_camera(dev);
	dev_gl860_camera_on(dev); // inversé avec reconf_cam

	// Init Isoc and URB
	err = usb_gl860_isoc_init(dev);

	if (err) {
		STK_ERROR("Failed to init ISOC stuff !\n");
		usb_gl860_isoc_cleanup(dev);
		gl860_free_buffers(dev);
		return err;
	}

	dev_gl860_init_camera(dev);

	// Start the video stream
	dev_gl860_start_stream(dev);

	// Video settings
	dev_gl860_camera_settings(dev);
*/
	dev->vopen++;
	fp->private_data = vdev;

	for (i=0; i<WCAM_NBR_SIZES; i++)
	{
		if (wcam_image_sizes[i].x ==  800 && _0503SIM_) { wcam_image_sizes[i].x=1280; wcam_image_sizes[i].y=960; }
		if (wcam_image_sizes[i].x == 1280 && (_0503A_ || _0503B_ || _0503C_)) wcam_image_sizes[i].y=1024;

		if (wcam_image_sizes[i].x ==  640 &&  nbVGA == 0)  nbVGA = i;
		if (wcam_image_sizes[i].x == 1280 && nbSXGA == 0) nbSXGA = i;
		if (wcam_image_sizes[i].x == 1600 && nbUXGA == 0) nbUXGA = i;
	}
	v4l_gl860_select_video_mode(dev, 640, 480);

	return 0;
}


/** 
 * @param inode Pointer on inode
 * @param fp File pointer
 * 
 * @returns 0 if all is OK
 *
 * @brief Release an opened file.
 *
 * This function permits to release an opened file with the 'open' method.
 */
static int v4l_gl860_release(struct inode *inode, struct file *fp)
{
	struct usb_gl860 *dev;
	struct video_device *vdev;
	
	vdev = video_devdata(fp);
	dev = video_get_drvdata(video_devdata(fp));

	dev->en_fct=0;

	if (dev->vopen == 0)
		STK_ERROR("v4l_release called on closed device\n");

	// Switch off the camera // Après "usb_gl860_isoc_cleanup(dev);" à l'origine
	dev_gl860_camera_off(dev);

	// Stop the video stream
	dev_gl860_stop_stream(dev);

	// ISOC and URB cleanup
	usb_gl860_isoc_cleanup(dev);

	// Free memory
	gl860_free_buffers(dev);

	dev_gl860_camera_asleep(dev);

	dev->vopen--;

	return 0;
}


/** 
 * @param fp File pointer
 *
 * @retval buf Buffer in user space
 * @retval count 
 * @retval f_pos 
 * 
 * @returns Count value
 *
 * @brief Read the video device
 *
 * This function is called by the application is reading the video device.
 */
static ssize_t v4l_gl860_read(struct file *fp, char __user *buf,
		size_t count, loff_t *f_pos)
{
	int noblock = fp->f_flags & O_NONBLOCK;

	struct usb_gl860 *dev;
	struct video_device *vdev;

	int bytes_to_read;
	void *image_buffer_addr;
	
	DECLARE_WAITQUEUE(wait, current);

	vdev = video_devdata(fp);
	dev = video_get_drvdata(video_devdata(fp));

	STK_STREAM("Read vdev=0x%p, buf=0x%p, count=%zd\n", vdev, buf, count);

	if (dev == NULL)
		return -EFAULT;

	if (vdev == NULL)
		return -EFAULT;

	if (dev->image_read_pos == 0) {
		add_wait_queue(&dev->wait_frame, &wait);

		while (dev->full_frames == NULL) {
			if (dev->error_status) {
				remove_wait_queue(&dev->wait_frame, &wait);
				set_current_state(TASK_RUNNING);
				return -dev->error_status ;
			}

			if (noblock) {
				remove_wait_queue(&dev->wait_frame, &wait);
				set_current_state(TASK_RUNNING);
				return -EWOULDBLOCK;
			}

			if (signal_pending(current)) {
				remove_wait_queue(&dev->wait_frame, &wait);
				set_current_state(TASK_RUNNING);
				return -ERESTARTSYS;
			}

			schedule();
			set_current_state(TASK_INTERRUPTIBLE);
		}

		remove_wait_queue(&dev->wait_frame, &wait);
		set_current_state(TASK_RUNNING);

		if (gl860_handle_frame(dev))
			return -EFAULT;
	}

	bytes_to_read = dev->view_size;

	if (dev->image_read_pos > bytes_to_read)
	{
		count = 0;
		dev->image_read_pos = bytes_to_read;
	}
	else
	{
		if (count + dev->image_read_pos > bytes_to_read)
			count = bytes_to_read - dev->image_read_pos;

		image_buffer_addr = dev->image_data;
		image_buffer_addr += dev->images[dev->fill_image].offset;
		image_buffer_addr += dev->image_read_pos;

		if (copy_to_user(buf, image_buffer_addr, count))
			return -EFAULT;
	
		dev->image_read_pos += count;
	}
	
	if (dev->image_read_pos >= bytes_to_read) {
		dev->image_read_pos = 0;
		gl860_next_image(dev);
	}

	return count;
}


/** 
 * @param fp File pointer
 * @param wait 
 * 
 * @returns 0 if all is OK
 *
 * @brief Polling function
 */
static unsigned int v4l_gl860_poll(struct file *fp, poll_table *wait)
{
	struct usb_gl860 *dev;
	struct video_device *vdev;
	
	vdev = video_devdata(fp);
	dev = video_get_drvdata(video_devdata(fp));

	STK_STREAM("Poll\n");

	if (vdev == NULL)
		return -EFAULT;

	if (dev == NULL)
		return -EFAULT;

	poll_wait(fp, &dev->wait_frame, wait);

	if (dev->error_status)
		return POLLERR;

	if (dev->full_frames != NULL)
		return (POLLIN | POLLRDNORM);

	return 0;
}


/** 
 * @param fp File pointer
 * @param vma VMA structure
 * 
 * @returns 0 if all is OK
 *
 * @brief Memory map
 *
 * This function permits to map a memory space.
 */
static int v4l_gl860_mmap(struct file *fp, struct vm_area_struct *vma)
{
	unsigned int i;

	unsigned long size;
	unsigned long start;
	unsigned long pos;
	unsigned long page;

	struct usb_gl860 *dev;

	struct video_device *vdev;
	
	vdev = video_devdata(fp);
	dev = video_get_drvdata(video_devdata(fp));

	STK_STREAM("mmap\n");

	start = vma->vm_start;
	size = vma->vm_end - vma->vm_start;

	// Find the buffer for this mapping...
	for (i=0; i<dev->nbuffers; i++) {
		pos = dev->images[i].offset;

		if ((pos >> PAGE_SHIFT) == vma->vm_pgoff)
			break;
	}

	// If no buffer found !
	if (i == WCAM_MAX_IMAGES) {
		STK_ERROR("mmap no buffer found !\n");
		return -EINVAL;
	}

	if (i == 0) {
		unsigned long total_size;

		total_size = dev->nbuffers * dev->len_per_image;

		if (size != dev->len_per_image && size != total_size)
			return -EINVAL;
	}
	else if (size > dev->len_per_image)
		return -EINVAL;

	vma->vm_flags |= VM_IO;

	pos = (unsigned long) dev->image_data;

	while (size > 0) {
		page = vmalloc_to_pfn((void *) pos);

		if (remap_pfn_range(vma, start, page, PAGE_SIZE, PAGE_SHARED))
			return -EAGAIN;

		start += PAGE_SIZE;
		pos += PAGE_SIZE;

		if (size > PAGE_SIZE)
			size -= PAGE_SIZE;
		else
			size = 0;
	}

	return 0;
}


/** 
 * @param inode Inode pointer
 * @param fp File pointer
 * @param cmd Command
 * @param arg Arguments of the command
 * 
 * @returns 0 if all is OK
 *
 * @brief Manage IOCTL
 *
 * This function permits to manage all the IOCTL from the application.
 */
static int v4l_gl860_do_ioctl(struct inode *inode, struct file *fp,
		unsigned int cmd, void __user *arg)
{
	struct usb_gl860 *dev;
	struct video_device *vdev;

	DECLARE_WAITQUEUE(wait, current);
	
	vdev = video_devdata(fp);
	dev = video_get_drvdata(video_devdata(fp));

#if (CONFIG_GL860_DEBUG == 1)
	v4l_printk_ioctl(cmd);
#endif
//STK_DEV("ioctl - En entruption ? %ld\n",in_interrupt());

	if ((cmd & 0xFFFF)==0x561C) { cmd=VIDIOC_S_CTRL; STK_DEBUG("Correction du cmd !!!\n"); } // xawtv ne préfixe pas bien sa commande !

	switch (cmd) {

		// Video 4 Linux v1

		case VIDIOCGCAP:
			{
				struct video_capability *cap = arg;

				STK_DEV("VIDIOCGCAP\n");

				memset(cap, 0, sizeof(*cap));
				strlcpy(cap->name, "gl860", sizeof(cap->name));
				cap->type = VID_TYPE_CAPTURE;
				cap->channels = 1;
				cap->audios = 0;

				switch (dev->webcam_type) {
					case WCAM_VGA:
						cap->minwidth = wcam_image_sizes[WCAM_80x60].x;
						cap->minheight = wcam_image_sizes[WCAM_80x60].y;
						cap->maxwidth = wcam_image_sizes[WCAM_640x480].x;
						cap->maxheight = wcam_image_sizes[WCAM_640x480].y;
						break;

					case WCAM_SXGA:
						cap->minwidth = wcam_image_sizes[WCAM_80x60].x;
						cap->minheight = wcam_image_sizes[WCAM_80x60].y;
						cap->maxwidth = wcam_image_sizes[WCAM_1280x960].x;
						cap->maxheight = wcam_image_sizes[WCAM_1280x960].y;
						break;

					case WCAM_UXGA:
						cap->minwidth = wcam_image_sizes[WCAM_80x60].x;
						cap->minheight = wcam_image_sizes[WCAM_80x60].y;
						cap->maxwidth = wcam_image_sizes[WCAM_1600x1200].x;
						cap->maxheight = wcam_image_sizes[WCAM_1600x1200].y;
						break;
				}
			}
			break;
	
		case VIDIOCGCHAN:
			{
			    struct video_channel *v = arg;

				STK_DEBUG("VIDIOCGCHAN\n");

			    if (v->channel != 0)
				    return -EINVAL;
			
				v->flags = 0;
			    v->tuners = 0;
			    v->type = VIDEO_TYPE_CAMERA;
			    strcpy(v->name, "Webcam");
			}
			break;

		case VIDIOCSCHAN:
			{
				struct video_channel *v = arg;

				STK_DEBUG("VIDIOCSCHAN\n");

				if (v->channel != 0)
					return -EINVAL;
			}
			break;

		case VIDIOCGPICT:
			{
				struct video_picture *p = arg;

				STK_DEBUG("VIDIOCGPICT\n");

				p->brightness = dev->vsettings.brightness;
				p->contrast = dev->vsettings.contrast;
				p->whiteness = dev->vsettings.whiteness;
				p->colour = dev->vsettings.saturation;
				p->depth = dev->vsettings.depth;
				p->palette = dev->vsettings.palette;
				p->hue = dev->vsettings.colour;

				switch (dev->vsettings.palette) {
					case WCAM_PALETTE_RGB24:
						p->palette = VIDEO_PALETTE_RGB24;
						break;

					case WCAM_PALETTE_RGB32:
						p->palette = VIDEO_PALETTE_RGB32;
						break;

					case WCAM_PALETTE_UYVY:
						p->palette = VIDEO_PALETTE_UYVY;
						break;

					case WCAM_PALETTE_YUYV:
						p->palette = VIDEO_PALETTE_YUYV;
						break;
				}
			}
			break;

		case VIDIOCSPICT:
			{
				struct video_picture *p = arg;
				dev->vsettings.v4l2=0;

				STK_DEBUG("VIDIOCSPICT\n");

				dev->vsettings.brightness = p->brightness;
				dev->vsettings.contrast = p->contrast;
				dev->vsettings.whiteness = p->whiteness;
				dev->vsettings.colour = p->hue;
				dev->vsettings.saturation = p->colour;
				
STK_DEV("La palette demandee VIDIOCSPICT : %d (r2_%d r3_%d u_%d y_%d)\n",p->palette,VIDEO_PALETTE_RGB24,VIDEO_PALETTE_RGB32,VIDEO_PALETTE_UYVY,VIDEO_PALETTE_YUYV);
				if (p->palette && p->palette != dev->vsettings.palette) {
					switch (p->palette) {
						case VIDEO_PALETTE_RGB24:
							dev->vsettings.depth = 24;
							dev->vsettings.palette = WCAM_PALETTE_RGB24;
							break;

						case VIDEO_PALETTE_RGB32:
							dev->vsettings.depth = 32;
							dev->vsettings.palette = WCAM_PALETTE_RGB32;
							break;

						case VIDEO_PALETTE_UYVY:
							dev->vsettings.depth = 16;
							dev->vsettings.palette = WCAM_PALETTE_UYVY;
							break;

						case VIDEO_PALETTE_YUYV:
							dev->vsettings.depth = 16;
							dev->vsettings.palette = WCAM_PALETTE_YUYV;
							break;

						default:
							return -EINVAL;
					}
				}

				dev_gl860_camera_settings(dev);

				STK_DEBUG("VIDIOCSPICT done\n");
			}
			break;

		case VIDIOCGWIN:
			{
				struct video_window *vw = arg;

				STK_DEBUG("VIDIOCGWIN\n");

				vw->x = dev->view.x;
				vw->y = dev->view.x;
				vw->width = dev->view.x;
				vw->height = dev->view.y;
				vw->chromakey = 0;
			}
			break;

		case VIDIOCSWIN:
			{
				struct video_window *vw = arg;

				STK_DEBUG("VIDIOCSWIN\n");

				STK_DEBUG("Set x=%d, y=%d\n", vw->x, vw->y);
				STK_DEBUG("Set width=%d, height=%d\n", vw->width, vw->height);
				STK_DEBUG("Flags = %X\n", vw->flags);

				if (dev->en_fct)
				{
					dev->en_fct=0;
					// Switch off the camera
					dev_gl860_camera_off(dev);

					// Stop the video stream
					dev_gl860_stop_stream(dev);
			
					// ISOC and URB cleanup
					usb_gl860_isoc_cleanup(dev);
	
					dev_gl860_camera_asleep(dev);
				}

				// Select the new video mode
				if (v4l_gl860_select_video_mode(dev, vw->width, vw->height)) {
					STK_ERROR("Select video mode failed !\n");
					return -EAGAIN;
				}

				// Clear the buffers
				gl860_clear_buffers(dev);

				// Initialize the device
				dev_gl860_init_camera(dev);
				dev_gl860_reconf_camera(dev);
				dev_gl860_camera_on(dev);

				// ISOC and URB init
				usb_gl860_isoc_init(dev);

				// Re-start the stream
				dev_gl860_start_stream(dev);

				// Video settings
				dev_gl860_camera_settings(dev);
				dev->en_fct=1;
			}
			break;

		case VIDIOCGFBUF:
			{
				struct video_buffer *vb = arg;

				STK_DEBUG("VIDIOCGFBUF\n");

				memset(vb, 0, sizeof(*vb));
			}
			break;

		case VIDIOCGMBUF:
			{
				int i;
				struct video_mbuf *vm = arg;

				STK_DEBUG("VIDIOCGMBUF\n");

				memset(vm, 0, sizeof(*vm));

				vm->size = dev->nbuffers * dev->len_per_image;
				vm->frames = dev->nbuffers;

				for (i=0; i<dev->nbuffers; i++)
					vm->offsets[i] = i * dev->len_per_image;
			}
			break;

		case VIDIOCMCAPTURE:
			{
				struct video_mmap *vm = arg;

				STK_DEBUG("VIDIOCMCAPTURE format=%d (r3_%d r2_%d u_%d y_%d)\n", vm->format,VIDEO_PALETTE_RGB32, VIDEO_PALETTE_RGB24,VIDEO_PALETTE_UYVY,VIDEO_PALETTE_YUYV);

				if (vm->frame < 0 || vm->frame >= dev->nbuffers)
					return -EINVAL;

				if (vm->format) {
					switch (vm->format) {
						case VIDEO_PALETTE_RGB32:
							break;

						case VIDEO_PALETTE_RGB24:
							break;

						case VIDEO_PALETTE_UYVY:
							break;

						case VIDEO_PALETTE_YUYV:
							break;

						default:
							return -EINVAL;
					}
				}

				if ((vm->width != dev->view.x) || (vm->height != dev->view.y)) 
					return -EAGAIN;

				if (dev->image_used[vm->frame])
					return -EBUSY;

				dev->image_used[vm->frame] = 1;

				STK_DEBUG("VIDIOCMCAPTURE done\n");
			}
			break;

		case VIDIOCSYNC:
			{
				int ret;
				int *mbuf = arg;

				STK_DEBUG("VIDIOCSYNC\n");

				if (*mbuf < 0 || *mbuf >= dev->nbuffers)
					return -EINVAL;

				if (dev->image_used[*mbuf] == 0)
					return -EINVAL;

				add_wait_queue(&dev->wait_frame, &wait);

				while (dev->full_frames == NULL) {
					if (dev->error_status) {
						remove_wait_queue(&dev->wait_frame, &wait);
						set_current_state(TASK_RUNNING);
						return -dev->error_status;
					}

					if (signal_pending(current)) {
						remove_wait_queue(&dev->wait_frame, &wait);
						set_current_state(TASK_RUNNING);
						return -ERESTARTSYS;
					}

					schedule();
					set_current_state(TASK_INTERRUPTIBLE);
				}

				remove_wait_queue(&dev->wait_frame, &wait);
				set_current_state(TASK_RUNNING);

				STK_DEBUG("VIDIOCSYNC: frame ready\n");

				dev->fill_image = *mbuf;

				ret = gl860_handle_frame(dev);

				if (ret != 0)
					STK_ERROR("VIDIOCSYNC error !\n");

				dev->image_used[*mbuf] = 0;
			}
			break;

		case VIDIOCGAUDIO:
			STK_DEBUG("VIDIOCGAUDIO\n");
			return -EINVAL;
			break;

		case VIDIOCSAUDIO:
			STK_DEBUG("VIDIOCSAUDIO\n");
			return -EINVAL;
			break;

		case VIDIOCGUNIT:
			{
				struct video_unit *vu = arg;

				vu->video = dev->vdev->minor & 0x3f;
				vu->audio = -1;
				vu->vbi = -1;
				vu->radio = -1;
				vu->teletext = -1;
			}
			break;


		// Video 4 Linux v2

		case VIDIOC_QUERYCAP:
			{
				struct v4l2_capability *cap = arg;

				STK_DEBUG("VIDIOC_QUERYCAP\n");

				memset(cap, 0, sizeof(*cap));
				strlcpy(cap->driver, "gl860", sizeof(cap->driver));

				cap->capabilities = V4L2_CAP_VIDEO_CAPTURE | V4L2_CAP_READWRITE | V4L2_CAP_STREAMING;
				cap->version = (__u32) DRIVER_VERSION_NUM, strlcpy(cap->card, dev->vdev->name, sizeof(cap->card));
			
				if (usb_make_path(dev->udev, cap->bus_info, sizeof(cap->bus_info)) < 0)
					strlcpy(cap->bus_info, dev->vdev->name, sizeof(cap->bus_info));
			}
			break;

		case VIDIOC_ENUMINPUT:
			{
				struct v4l2_input *i = arg;

				STK_DEBUG("VIDIOC_ENUMINPUT %d\n", i->index);

				if (i->index)
					return -EINVAL;

				strlcpy(i->name, "USB", sizeof(i->name));
				i->type = V4L2_INPUT_TYPE_CAMERA;
			}
			break;

		case VIDIOC_G_INPUT:
			{
				struct v4l2_input *i = arg;

				STK_DEBUG("GET INPUT %d\n", i->index);

				if (i->index)
					return -EINVAL;
			}
			break;

		case VIDIOC_S_INPUT:
			{
				struct v4l2_input *i = arg;

				STK_DEBUG("SET INPUT %d\n", i->index);

				if (i->index != 0)
					return -EINVAL;
			}
			break;

		case VIDIOC_QUERYCTRL:
			{
				int i;
				int nbr;
				struct v4l2_queryctrl *c = arg;
				int lez=(dev->webcam_model==GENESYS_0503)*0;

				STK_DEBUG("VIDIOC_QUERYCTRL id = %d\n", c->id);

				nbr = sizeof(gl860_controls)/sizeof(struct v4l2_queryctrl);//-(3*lez+2*(!lez));

				for (i=0; i<nbr; i++) {
					if (gl860_controls[i+3*lez].id == c->id) {
						STK_DEBUG("VIDIOC_QUERYCTRL found\n");
						memcpy(c, &gl860_controls[i+3*lez], sizeof(struct v4l2_queryctrl));
						break;
					}
				}

				if (i >= nbr)
					return -EINVAL;
			}
			break;

		case VIDIOC_G_CTRL:
			{
				struct v4l2_control *c = arg;

				STK_DEBUG("GET CTRL id=%d\n", c->id);

				switch (c->id) {
					case V4L2_CID_BRIGHTNESS:
						c->value = dev->vsettings.brightness;
						break;

					case V4L2_CID_WHITENESS:
						c->value = dev->vsettings.whiteness;
						break;

					case V4L2_CID_SATURATION:
						c->value = dev->vsettings.saturation;
						break;

					case V4L2_CID_CONTRAST:
						c->value = dev->vsettings.contrast;
						break;

					case V4L2_CID_HUE:
						c->value = dev->vsettings.colour;
						break;

					case V4L2_CID_SHARPNESS:
						c->value = dev->vsettings.net;
						break;

					case V4L2_CID_POWER_LINE_FREQUENCY:
						c->value = dev->vsettings.freq50;
						break;

					case V4L2_CID_BACKLIGHT_COMPENSATION:
						c->value = dev->vsettings.contrej;
						break;

					case V4L2_CID_SRC_LUM:
						c->value = dev->vsettings.srclum;
						break;

					case V4L2_CID_HFLIP:
						c->value = dev->vsettings.miroirh;
						break;

					case V4L2_CID_VFLIP:
						c->value = dev->vsettings.miroirv;
						break;

					default:
						return -EINVAL;
				}
			}
			break;

		case VIDIOC_S_CTRL:
			{
				struct v4l2_control *c = arg;

				STK_DEBUG("SET CTRL id=%d value=%d\n", c->id, c->value);  // utilisé par xawtv

				switch (c->id) {
					case V4L2_CID_BRIGHTNESS:
						dev->vsettings.brightness = (0xff00 & c->value);
						break;

					case V4L2_CID_WHITENESS:
						dev->vsettings.whiteness = (0xff00 & c->value);
						break;

					case V4L2_CID_SATURATION:
						dev->vsettings.saturation = (0xff00 & c->value);
						break;

					case V4L2_CID_CONTRAST:
						dev->vsettings.contrast = (0xff00 & c->value);
						break;

					case V4L2_CID_HUE:
						dev->vsettings.colour=(0xff00 & c->value);
						break;

					case V4L2_CID_SHARPNESS:
						dev->vsettings.net=(0xff00 & c->value);
						break;

					case V4L2_CID_POWER_LINE_FREQUENCY:
						dev->vsettings.freq50=c->value;
						break;

					case V4L2_CID_BACKLIGHT_COMPENSATION:
						dev->vsettings.contrej=(0xff00 & c->value);
						break;

					case V4L2_CID_SRC_LUM:
						dev->vsettings.srclum=(0xff00 & c->value);
						break;

					case V4L2_CID_HFLIP:
						dev->vsettings.miroirh=c->value;
						break;

					case V4L2_CID_VFLIP:
						dev->vsettings.miroirv=c->value;
						break;

					default:
						return -EINVAL;
				}

				dev_gl860_camera_settings(dev);
			}
			break;

		case VIDIOC_ENUM_FMT:
			{
				int index;
				struct v4l2_fmtdesc *fmtd = arg;

				STK_DEV("VIDIOC_ENUM_FMT %d\n", fmtd->index);

				if (fmtd->index > 6)//!= 0)
					return -EINVAL;

				index = fmtd->index;

				memset(fmtd, 0, sizeof(*fmtd));

				fmtd->type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
				fmtd->index = index;

				switch (index) {
/*
					case 0:
						fmtd->flags = 0;
						fmtd->pixelformat = V4L2_PIX_FMT_SGBRG8;

						strcpy(fmtd->description, "srggb8");
						break;
*/

					case 0:
						fmtd->flags = 0;
						fmtd->pixelformat = V4L2_PIX_FMT_UYVY;

						strcpy(fmtd->description, "uyvy");
						break;

					case 1:
						fmtd->flags = 0;
						fmtd->pixelformat = V4L2_PIX_FMT_YUYV;

						strcpy(fmtd->description, "yuyv");
						break;

					case 2:
						fmtd->flags = 0;
						fmtd->pixelformat = V4L2_PIX_FMT_RGB8;

						strcpy(fmtd->description, "rgb8");
						break;

/*
					case 0:
						fmtd->flags = 0;
						fmtd->pixelformat = V4L2_PIX_FMT_RGB24;

						strcpy(fmtd->description, "rgb24");
						break;

					case 1:
						fmtd->flags = 0;
						fmtd->pixelformat = V4L2_PIX_FMT_RGB32;

						strcpy(fmtd->description, "rgb32");
						break;

					case 2:
						fmtd->flags = 0;
						fmtd->pixelformat = V4L2_PIX_FMT_BGR24;

						strcpy(fmtd->description, "bgr24");
						break;

					case 3:
						fmtd->flags = 0;
						fmtd->pixelformat = V4L2_PIX_FMT_BGR32;

						strcpy(fmtd->description, "bgr32");
						break;

					case 4:
						fmtd->flags = 0;
						fmtd->pixelformat = V4L2_PIX_FMT_UYVY;

						strcpy(fmtd->description, "uyvy");
						break;

					case 5:
						fmtd->flags = 0;
						fmtd->pixelformat = V4L2_PIX_FMT_YUYV;

						strcpy(fmtd->description, "yuyv");
						break;

					case 6:
						fmtd->flags = 0;
						fmtd->pixelformat = V4L2_PIX_FMT_RGB8;

						strcpy(fmtd->description, "rgb8");
						break;
*/
					default:
						return -EINVAL;
				}
			}
			break;

		case VIDIOC_ENUM_FRAMESIZES:
			{
				int index;
				struct v4l2_frmsizeenum *fsize = arg;

				index = fsize->index;
				STK_DEV("VIDIOC_ENUM_FRAMESIZES %d\n", index);

				if (dev->webcam_type==WCAM_VGA && index>nbVGA) return -EINVAL;
				if (dev->webcam_type==WCAM_SXGA && index>nbSXGA) return -EINVAL;
				if (dev->webcam_type==WCAM_UXGA && index>nbUXGA) return -EINVAL;

				fsize->type = V4L2_FRMSIZE_TYPE_DISCRETE;
				fsize->discrete.width = wcam_image_sizes[index].x;
				fsize->discrete.height = wcam_image_sizes[index].y;
			}
			break;

		case VIDIOC_G_FMT:
			{
				struct v4l2_format *fmtd = arg;
				struct v4l2_pix_format pix_format;

				STK_DEV("GET FMT %d (VID_CAP=%d)\n", fmtd->type, V4L2_BUF_TYPE_VIDEO_CAPTURE);

				if (fmtd->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
					return -EINVAL;

				pix_format.width = dev->view.x;
				pix_format.height = dev->view.y;
				pix_format.field = V4L2_FIELD_NONE;
				pix_format.colorspace = V4L2_COLORSPACE_SRGB;
				pix_format.priv = 0;

				switch (dev->vsettings.palette) {
					case WCAM_PALETTE_GRIS:
						pix_format.pixelformat = V4L2_PIX_FMT_SGBRG8;
						//pix_format.pixelformat = V4L2_PIX_FMT_RGB8;
						pix_format.sizeimage = pix_format.width * pix_format.height * 1;
						pix_format.bytesperline = 1 * pix_format.width;
						break;

					case WCAM_PALETTE_RGB24:
						pix_format.pixelformat = V4L2_PIX_FMT_RGB24;
						pix_format.sizeimage = pix_format.width * pix_format.height * 3;
						pix_format.bytesperline = 3 * pix_format.width;
						break;

					case WCAM_PALETTE_RGB32:
						pix_format.pixelformat = V4L2_PIX_FMT_RGB32;
						pix_format.sizeimage = pix_format.width * pix_format.height * 4;
						pix_format.bytesperline = 4 * pix_format.width;
						break;

					case WCAM_PALETTE_BGR24:
						pix_format.pixelformat = V4L2_PIX_FMT_BGR24;
						pix_format.sizeimage = pix_format.width * pix_format.height * 3;
						pix_format.bytesperline = 3 * pix_format.width;
						break;

					case WCAM_PALETTE_BGR32:
						pix_format.pixelformat = V4L2_PIX_FMT_BGR32;
						pix_format.sizeimage = pix_format.width * pix_format.height * 4;
						pix_format.bytesperline = 4 * pix_format.width;
						break;

					case WCAM_PALETTE_UYVY:
						pix_format.pixelformat = V4L2_PIX_FMT_UYVY;
						pix_format.sizeimage = pix_format.width * pix_format.height * 2;
						pix_format.bytesperline = 2 * pix_format.width;
						break;

					case WCAM_PALETTE_YUYV:
						pix_format.pixelformat = V4L2_PIX_FMT_YUYV;
						pix_format.sizeimage = pix_format.width * pix_format.height * 2;
						pix_format.bytesperline = 2 * pix_format.width;
						break;
				}

				memcpy(&(fmtd->fmt.pix), &pix_format, sizeof(pix_format));
			}
			break;

		case VIDIOC_TRY_FMT:
			{
				struct v4l2_format *fmtd = arg;

				STK_DEV("TRY FMT %d (VID_CAP=%d)\n", fmtd->type, V4L2_BUF_TYPE_VIDEO_CAPTURE);

				if (fmtd->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
					return -EINVAL;

				switch (fmtd->fmt.pix.pixelformat) {
/*					case V4L2_PIX_FMT_RGB24:
					case V4L2_PIX_FMT_BGR24:
						dev->vsettings.depth = 24;
						break;
*/
					case V4L2_PIX_FMT_RGB8:
					case V4L2_PIX_FMT_SGBRG8:
						dev->vsettings.depth = 8;
						break;
/*
					case V4L2_PIX_FMT_RGB32:
					case V4L2_PIX_FMT_BGR32:
						dev->vsettings.depth = 32;
						break;
*/
					case V4L2_PIX_FMT_UYVY:
					case V4L2_PIX_FMT_YUYV:
						dev->vsettings.depth = 16;
						break;

					default:
						return -EINVAL;
				}

				switch (dev->webcam_type) {
					case WCAM_VGA:
						if (fmtd->fmt.pix.width > wcam_image_sizes[nbVGA].x)
							 fmtd->fmt.pix.width = wcam_image_sizes[nbVGA].x;
						else if (fmtd->fmt.pix.width < wcam_image_sizes[0].x)
							fmtd->fmt.pix.width = wcam_image_sizes[0].x;
	
						if (fmtd->fmt.pix.height > wcam_image_sizes[nbVGA].y)
							 fmtd->fmt.pix.height = wcam_image_sizes[nbVGA].y;
						else if (fmtd->fmt.pix.height < wcam_image_sizes[0].y)
							fmtd->fmt.pix.height = wcam_image_sizes[0].y;
						break;

					case WCAM_SXGA:
						if (fmtd->fmt.pix.width > wcam_image_sizes[nbSXGA].x)
							 fmtd->fmt.pix.width = wcam_image_sizes[nbSXGA].x;
						else if (fmtd->fmt.pix.width < wcam_image_sizes[0].x)
							fmtd->fmt.pix.width = wcam_image_sizes[0].x;
	
						if (fmtd->fmt.pix.height > wcam_image_sizes[nbSXGA].y)
							 fmtd->fmt.pix.height = wcam_image_sizes[nbSXGA].y;
						else if (fmtd->fmt.pix.height < wcam_image_sizes[0].y)
							fmtd->fmt.pix.height = wcam_image_sizes[0].y;
						break;

					case WCAM_UXGA:
						if (fmtd->fmt.pix.width > wcam_image_sizes[nbUXGA].x)
							 fmtd->fmt.pix.width = wcam_image_sizes[nbUXGA].x;
						else if (fmtd->fmt.pix.width < wcam_image_sizes[0].x)
							fmtd->fmt.pix.width = wcam_image_sizes[0].x;
	
						if (fmtd->fmt.pix.height > wcam_image_sizes[nbUXGA].y)
							 fmtd->fmt.pix.height = wcam_image_sizes[nbUXGA].y;
						else if (fmtd->fmt.pix.height < wcam_image_sizes[0].y)
							fmtd->fmt.pix.height = wcam_image_sizes[0].y;
						break;
				}

			}
			break;

		case VIDIOC_S_FMT:
			{
				struct v4l2_format *fmtd = arg;

				STK_DEV("SET FMT %d (VID_CAP=%d): %04x (r23_%04x %04x - b23_%04x %04x - uy_%04x %04x)\n", 
					fmtd->type, V4L2_BUF_TYPE_VIDEO_CAPTURE, fmtd->fmt.pix.pixelformat, 
					V4L2_PIX_FMT_RGB24, V4L2_PIX_FMT_RGB32, V4L2_PIX_FMT_BGR24, V4L2_PIX_FMT_BGR32,
				   V4L2_PIX_FMT_UYVY, V4L2_PIX_FMT_YUYV
					);
				if (fmtd->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
					return -EINVAL;

				switch (fmtd->fmt.pix.pixelformat) {
					case V4L2_PIX_FMT_RGB8:
					case V4L2_PIX_FMT_SGBRG8:
						dev->vsettings.depth = 8;
						dev->vsettings.palette = WCAM_PALETTE_GRIS;
						break;

					case V4L2_PIX_FMT_RGB24:
						dev->vsettings.depth = 24;
						dev->vsettings.palette = WCAM_PALETTE_RGB24;
						break;

					case V4L2_PIX_FMT_RGB32:
						dev->vsettings.depth = 32;
						dev->vsettings.palette = WCAM_PALETTE_RGB32;
						break;

					case V4L2_PIX_FMT_BGR24:
						dev->vsettings.depth = 24;
						dev->vsettings.palette = WCAM_PALETTE_BGR24;
						break;

					case V4L2_PIX_FMT_BGR32:
						dev->vsettings.depth = 32;
						dev->vsettings.palette = WCAM_PALETTE_BGR32;
						break;

					case V4L2_PIX_FMT_UYVY:
						dev->vsettings.depth = 16;
						dev->vsettings.palette = WCAM_PALETTE_UYVY;
						break;

					case V4L2_PIX_FMT_YUYV:
						dev->vsettings.depth = 16;
						dev->vsettings.palette = WCAM_PALETTE_YUYV;
						break;

					default:
						return -EINVAL;
				}
				dev->vsettings.v4l2=1;

				STK_DEBUG("Set width=%d, height=%d\n", fmtd->fmt.pix.width, fmtd->fmt.pix.height);

				if (dev->en_fct)
				{
					dev->en_fct=0;
					// Switch off the camera
					dev_gl860_camera_off(dev);

					// Stop the video stream
					dev_gl860_stop_stream(dev);
				
					// ISOC and URB cleanup
					usb_gl860_isoc_cleanup(dev);

					dev_gl860_camera_asleep(dev);
				}

				// Select the new video mode
				if (v4l_gl860_select_video_mode(dev, fmtd->fmt.pix.width, fmtd->fmt.pix.height)) {
					STK_ERROR("Select video mode failed !\n");
					return -EAGAIN;
				}

				// Clear the buffers
				gl860_clear_buffers(dev);

				// Initialize the device
				dev_gl860_init_camera(dev);
				dev_gl860_reconf_camera(dev);
				dev_gl860_camera_on(dev);

				// ISOC and URB init
				usb_gl860_isoc_init(dev);

				// Re-start the stream
				dev_gl860_start_stream(dev);

				// Video settings
				dev_gl860_camera_settings(dev);
				dev->en_fct=1;
			}
			break;

		case VIDIOC_QUERYSTD:
			{
				STK_DEBUG("QUERY STD\n");
				return -EINVAL;
			}
			break;

		case VIDIOC_G_STD:
			{
				v4l2_std_id *std = arg;

				STK_DEBUG("GET STD\n");
		
				*std = V4L2_STD_UNKNOWN;
			}
			break;

		case VIDIOC_S_STD:
			{
				v4l2_std_id *std = arg;

				STK_DEBUG("SET STD\n");
				
				if (*std != V4L2_STD_UNKNOWN)
					return -EINVAL;
			}
			break;

		case VIDIOC_ENUMSTD:
			{
				struct v4l2_standard *std = arg;

				STK_DEBUG("VIDIOC_ENUMSTD\n");

				if (std->index != 0)
					return -EINVAL;

				std->id = V4L2_STD_UNKNOWN;
				strncpy(std->name, "webcam", sizeof(std->name));

				break;
			}

		case VIDIOC_REQBUFS:
			{
				int nbuffers;
				struct v4l2_requestbuffers *rb = arg;

				if (rb->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
					return -EINVAL;

				if (rb->memory != V4L2_MEMORY_MMAP)
					return -EINVAL;

				nbuffers = rb->count;

				if (nbuffers < 2)
					nbuffers = 2;
				else if (nbuffers > dev->nbuffers)
					nbuffers = dev->nbuffers;

				rb->count = dev->nbuffers;
			}
			break;

		case VIDIOC_QUERYBUF:
			{
				int index;
				struct v4l2_buffer *buf = arg;

				STK_DEBUG("QUERY BUFFERS %d %d\n", buf->index, dev->nbuffers);

				if (buf->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
					return -EINVAL;

				if (buf->memory != V4L2_MEMORY_MMAP) 
					return -EINVAL;

				index = buf->index;

				if (index < 0 || index >= dev->nbuffers)
					return -EINVAL;

				memset(buf, 0, sizeof(struct v4l2_buffer));

				buf->type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
				buf->index = index;
				buf->m.offset = index * dev->len_per_image;
				buf->bytesused = dev->image_size;
				buf->field = V4L2_FIELD_NONE;
				buf->memory = V4L2_MEMORY_MMAP;
				buf->length = dev->len_per_image;
			}
			break;

		case VIDIOC_QBUF:
			{
				struct v4l2_buffer *buf = arg;

				STK_DEBUG("VIDIOC_QBUF\n");

				if (buf->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
					return -EINVAL;

				if (buf->memory != V4L2_MEMORY_MMAP)
					return -EINVAL;

				if (buf->index < 0 || buf->index >= dev->nbuffers)
					return -EINVAL;

				buf->flags |= V4L2_BUF_FLAG_QUEUED;
				buf->flags &= ~V4L2_BUF_FLAG_DONE;
			}
			break;

		case VIDIOC_DQBUF:
			{
				int ret;
				struct v4l2_buffer *buf = arg;

				STK_DEBUG("VIDIOC_DQBUF\n");
				
				if (buf->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
					return -EINVAL;

				add_wait_queue(&dev->wait_frame, &wait);

				while (dev->full_frames == NULL) {
					if (dev->error_status) {
						remove_wait_queue(&dev->wait_frame, &wait);
						set_current_state(TASK_RUNNING);

						return -dev->error_status;
					}

					if (signal_pending(current)) {
						remove_wait_queue(&dev->wait_frame, &wait);
						set_current_state(TASK_RUNNING);

						return -ERESTARTSYS;
					}

					schedule();
					set_current_state(TASK_INTERRUPTIBLE);
				}

				remove_wait_queue(&dev->wait_frame, &wait);
				set_current_state(TASK_RUNNING);

				STK_DEV("VIDIOC_DQBUF : frame ready.\n"); // À chaque image on passe par là

				ret = gl860_handle_frame(dev);

				if (ret)
					return -EFAULT;

				buf->index = dev->fill_image;
				buf->bytesused = dev->image_size;
				buf->flags = V4L2_BUF_FLAG_MAPPED;
				buf->field = V4L2_FIELD_NONE;
				do_gettimeofday(&buf->timestamp);
				buf->sequence = 0;
				buf->memory = V4L2_MEMORY_MMAP;
				buf->m.offset = dev->fill_image * dev->len_per_image;
				buf->length = buf->bytesused;

				gl860_next_image(dev);
			}
			break;

		case VIDIOC_STREAMON:
			{
				STK_DEBUG("VIDIOC_STREAMON\n");

				usb_gl860_isoc_init(dev);
			}
			break;

		case VIDIOC_STREAMOFF:
			{
				STK_DEBUG("VIDIOC_STREAMOFF\n");

				usb_gl860_isoc_cleanup(dev);
			}
			break;

		case VIDIOC_G_PARM:
			{
				struct v4l2_streamparm *sp = arg;

				STK_DEBUG("GET PARM %d\n", sp->type);

				if (sp->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
					return -EINVAL;

				sp->parm.capture.capability = 0;
				sp->parm.capture.capturemode = 0;
				sp->parm.capture.timeperframe.numerator = 1;
				sp->parm.capture.timeperframe.denominator = 30;
				sp->parm.capture.readbuffers = 2;
				sp->parm.capture.extendedmode = 0;
			}
			break;


		case VIDIOC_G_AUDIO:
			STK_DEBUG("GET AUDIO\n");
			return -EINVAL;
			break;

		case VIDIOC_S_AUDIO:
			STK_DEBUG("SET AUDIO\n");
			return -EINVAL;
			break;

		case VIDIOC_S_TUNER:
			STK_DEBUG("SET TUNER\n");
			return -EINVAL;
			break;

		case VIDIOCGTUNER:
		case VIDIOC_G_FBUF:
		case VIDIOC_S_FBUF:
		case VIDIOC_OVERLAY:
			return -EINVAL;
			break;

		case VIDIOC_G_TUNER:
		case VIDIOC_G_FREQUENCY:
		case VIDIOC_S_FREQUENCY:
			return -EINVAL;
			break;

		case VIDIOC_QUERYMENU:
			return -EINVAL;
			break;
/*
		case VIDIOC_CROPCAP:
			{
				struct v4l2_cropcap cc;

				cc.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
				cc.pixelaspect.numerator = 1;
				cc.pixelaspect.denominator = 1;
				cc.bounds.top = 0;
				cc.bounds.left = 0;
				cc.bounds.width = 640;
				cc.bounds.height = 480;
				cc.defrect.top = 0;
				cc.defrect.left = 0;
				cc.defrect.width = 640;
				cc.defrect.height = 480;

				memcpy(arg, &cc, sizeof(cc));
			}
			break;
*/
		default:
			STK_DEBUG("IOCTL unknown !\n");
			return -ENOIOCTLCMD;
	}

	return 0;
}


/** 
 * @param inode Inode pointer
 * @param fp File pointer
 * @param cmd Command
 * @param arg Arguements of the command
 * 
 * @returns 0 if all is OK
 *
 * @brief Manage IOCTL
 *
 * This function permits to manage all the IOCTL from the application.
 */
static int v4l_gl860_ioctl(struct inode *inode, struct file *fp,
		unsigned int cmd, unsigned long arg)
{
	int err;
	struct usb_gl860 *dev;
	struct video_device *vdev;
	
	vdev = video_devdata(fp);
	dev = video_get_drvdata(video_devdata(fp));

	STK_DEBUG("v4l_gl860_ioctl %02X\n", (unsigned char) cmd);

	if (dev == NULL)
		return -EFAULT;

	if (vdev == NULL)
		return -EFAULT;

	err = v4l_gl860_do_ioctl(inode, fp, cmd, (void __user *) arg);

	return err;
}


/** 
 * @param dev Device structure
 * 
 * @returns 0 if all is OK
 *
 * @brief Register the video device
 *
 * This function permits to register the USB device to the video device.
 */
int v4l_gl860_register_video_device(struct usb_gl860 *dev)
{
	int err;

	strcpy(dev->vdev->name, DRIVER_DESC);

//	dev->vdev->vfl_type = VID_TYPE_CAPTURE;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
	dev->vdev->dev = &dev->interface->dev;
	dev->vdev->owner = THIS_MODULE;
	dev->vdev->type = VID_TYPE_CAPTURE;
	dev->vdev->hardware = VID_HARDWARE_GL860;
#else
	dev->vdev->parent = &(dev->interface->dev);
#endif
	dev->vdev->fops = &v4l_gl860_fops;
	dev->vdev->release = video_device_release;
	dev->vdev->minor = -1;

	video_set_drvdata(dev->vdev, dev);

	err = video_register_device(dev->vdev, VFL_TYPE_GRABBER, -1);

	if (err)
		STK_ERROR("Video register fail !\n");
	else
		STK_INFO("Genesys USB2.0 Camera is now controlling video device /dev/video%d\n", dev->vdev->minor);

	return err;
}


/** 
 * @param dev Device structure
 * 
 * @returns 0 if all is OK
 *
 * @brief Unregister the video device
 *
 * This function permits to unregister the video device.
 */
int v4l_gl860_unregister_video_device(struct usb_gl860 *dev)
{
	STK_INFO("Genesys USB2.0 Camera release resources video device /dev/video%d\n", dev->vdev->minor);

	video_set_drvdata(dev->vdev, NULL);
	video_unregister_device(dev->vdev);

	return 0;
}


/**
 * @var v4l_gl860_fops
 *
 * This variable contains some callback
 */
static struct file_operations v4l_gl860_fops = {
	.owner = THIS_MODULE,
	.open = v4l_gl860_open,
	.release = v4l_gl860_release,
	.read = v4l_gl860_read,
	.poll = v4l_gl860_poll,
	.mmap = v4l_gl860_mmap,
	.ioctl = v4l_gl860_ioctl,
#ifdef CONFIG_COMPAT
	.compat_ioctl = v4l_compat_ioctl32,
#endif
	.llseek = no_llseek
};

