/** 
 * @file gl860.h
 * @author Olivier LORIN, tiré du pilote Syntek par Nicolas VIVIEN
 * @date 2008-05-05
 * @version v0.9.0
 *
 * @brief Driver for GL860 USB video camera
 *
 * @note Copyright (C) Olivier LORIN
 *
 * @par Licences
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#ifndef GL860_H
#define GL860_H


#define DRIVER_NAME					"gl860"					/**< Name of this driver */
#define DRIVER_VERSION				"v0.3.0"					/**< Version of this driver */
#define DRIVER_VERSION_NUM			0x000300					/**< Version numerical of this driver */
#define DRIVER_DESC					"Genesys gl860 USB Video Camera"	/**< Short description of this driver */
#define DRIVER_AUTHOR				"Nol"			/**< Author of this driver */
#define PREFIX						DRIVER_NAME ": "			/**< Prefix use for the STK "printk" */

#define USB_GENESYS_VENDOR_ID		0x05e3						/**< Vendor ID of the camera */

#define USB_F191_PRODUCT_ID		0xf191						/**< Product ID of the camera */
#define USB_0503_PRODUCT_ID		0x0503						/**< Product ID of the camera */

#define IDF191    1
#define ID0503MS  2
#define ID0503SIM 4
#define ID0503HS  8
#define ID0503IM  16

#define  _0503SIM_ (dev->pilote==ID0503SIM)
#define  _0503MS_  (dev->pilote==ID0503MS)
#define  _0503C_   (dev->pilote==ID0503HS)
#define  _0503B_   0
#define  _0503A_   (dev->pilote==ID0503IM)
#define  _F191_    (dev->pilote==IDF191)

// Libellés des paramètres

#ifdef _FRANCAIS_
   #define LBL_MIROIR_H   "Miroir Hrz"
   #define LBL_MIROIR_V   "Miroir Vrt"
   #define LBL_LUMINOSITE "Luminosite"
   #define LBL_NETTETE    "Nettete"
   #define LBL_CONTREJ    "Contre-jour"
   #define LBL_SRC_LUM    "Source"
   #define LBL_CONTRASTE  "Contraste"
   #define LBL_BLANCHEUR  "Blancheur"
#else
   #define LBL_MIROIR_H   "Mirror"
   #define LBL_MIROIR_V   "Flip"
   #define LBL_LUMINOSITE "Brightness"
   #define LBL_NETTETE    "Sharpness"
   #define LBL_CONTREJ    "Backlight"
   #define LBL_SRC_LUM    "Light Source"
   #define LBL_CONTRASTE  "Contrast"
   #define LBL_BLANCHEUR  "Whiteness"
#endif

/**
 * @def VID_HARDWARE_GL860
 *
 * This value must be inserted into the kernel headers linux/videodev.h
 * It's useful only for the support of V4L v1
 */
#define VID_HARDWARE_GL860		88


/**
 * @def MAX_ISO_BUFS
 *   Number maximal of ISOC buffers
 *
 * @def ISO_FRAMES_PER_DESC
 *   Number frames per ISOC descriptor
 *
 * @def ISO_MAX_FRAME_SIZE
 *   Maximale size of frame
 *
 * @def ISO_BUFFER_SIZE
 *   Maximal size of buffer
 */
#define MAX_ISO_BUFS					2
#define ISO_FRAMES_PER_DESC		32
//#define ISO_MAX_FRAME_SIZE			2048
//#define ISO_BUFFER_SIZE				(ISO_FRAMES_PER_DESC * ISO_MAX_FRAME_SIZE)


/**
 * @def GL860_MAX_IMAGES
 *   Absolute maximum number of buffers available for mmap()
 *
 * @def GL860_FRAME_SIZE
 *   Maximum size after decompression
 */
#define WCAM_MAX_IMAGES			10
#define WCAM_FRAME_SIZE			(1600 * 1200 * 4)







/**
 * @def DRIVER_SUPPORT
 * 
 * List of supported device
 *   - DC1125 : USB2.0 Syntek chipset
 */
#define DRIVER_SUPPORT				"Genesys USB Camera : gl860"

/**
 * @def CONFIG_GL860_DEBUG
 *   Enable / Disable the debug mode.
 *
 * @def STK_INFO(str, args...)
 *   Print information message. 
 *   @a Use this function like the function printf.
 *
 * @def STK_ERROR(str, args...)
 *   Print error message.
 *   @a Use this function like the function printf.
 *
 * @def STK_WARNING(str, args...)
 *   Print warning message.
 *   @a Use this function like the function printf.
 *
 * @def STK_DEBUG(str, args...)
 *   Print debug message.
 *   @a Use this function like the function printf.
 */
#ifndef CONFIG_GL860_DEBUG
#define CONFIG_GL860_DEBUG 			0
#endif

//#undef CONFIG_GL860_DEBUG_STREAM
//#define CONFIG_GL860_DEBUG_STREAM 1
//#undef CONFIG_GL860_DEBUG
//#define CONFIG_GL860_DEBUG 1

#if CONFIG_GL860_DEBUG

#define STK_INFO(str, args...)			printk(KERN_INFO PREFIX str, ##args)
#define STK_ERROR(str, args...)			printk(KERN_ERR PREFIX str, ##args)
#define STK_WARNING(str, args...)		printk(KERN_WARNING PREFIX str, ##args)
#define STK_DEBUG(str, args...)			printk(KERN_DEBUG PREFIX str, ##args)

#else

#define STK_INFO(str, args...)			printk(KERN_INFO PREFIX str, ##args)
#define STK_ERROR(str, args...)			printk(KERN_ERR PREFIX str, ##args)
#define STK_WARNING(str, args...)		printk(KERN_WARNING PREFIX str, ##args)
#define STK_DEBUG(str, args...)			do { } while(0)

#endif

//#define STK_DEV(str, args...)			printk(KERN_INFO PREFIX str, ##args)
#define STK_DEV(str, args...)				do { } while(0)

/**
 * @def CONFIG_GL860_DEBUG_STREAM
 *   Enable / Disable the debug mode about the stream.
 *
 * @def STK_STREAM(str, args...)
 *   Print stream debug message. 
 *   @a Use this function like the function printf.
 */
#ifndef CONFIG_GL860_DEBUG_STREAM
#define CONFIG_GL860_DEBUG_STREAM		0
#endif

#if CONFIG_GL860_DEBUG_STREAM

#define STK_STREAM(str, args...)		printk(KERN_DEBUG PREFIX str, ##args)

#else

#define STK_STREAM(str, args...)		do { } while(0)

#endif



/**
 * @enum T_SYNTEK_DEVICE Video camera supported by the driver
 */
typedef enum {
	GENESYS_F191 = 1,
	GENESYS_0503 = 2,
} T_SYNTEK_DEVICE;


/**
 * @enum T_GL860_VIDEOMODE Video feature supported by camera
 */
typedef enum {
	WCAM_VGA,						/**< For VGA video camera */
	WCAM_SXGA,						/**< For SXGA video camera 1.3M */
	WCAM_UXGA,						/**< For UXGA video camera 2M */
} T_WCAM_VIDEOMODE;


/** 
 * @enum T_GL860_RESOLUTION Video resolution
 */
typedef enum {
	WCAM_80x60,
	WCAM_128x96,
	WCAM_160x120,
	WCAM_213x160,
	WCAM_320x240,
	WCAM_640x480,
	WCAM_800x600,
	WCAM_1280x960,
	WCAM_1600x1200,
	WCAM_NBR_SIZES
} T_WCAM_RESOLUTION;


/**
 * @enum T_GL860_PALETTE Color palette
 */
typedef enum {
	WCAM_PALETTE_RGB24,
	WCAM_PALETTE_RGB32,
	WCAM_PALETTE_BGR24,
	WCAM_PALETTE_BGR32,
	WCAM_PALETTE_UYVY,
	WCAM_PALETTE_YUYV,
	WCAM_PALETTE_GRIS
} T_WCAM_PALETTE;

#define V4L2_PIX_FMT_RGB8 0x20203859
//'  8Y'

/**
 * @struct gl860_iso_buf
 */
struct gl860_iso_buf {
	void *data;
	int length;
	int read;
	struct urb *urb;
};


/**
 * @struct gl860_frame_buf
 */
struct gl860_frame_buf {
	int errors;
	void *data;
	volatile int filled;
	struct gl860_frame_buf *next;
};


/**
 * @struct gl860_image_buf
 */
struct gl860_image_buf {
	unsigned long offset;				/**< Memory offset */
	int vma_use_count;					/**< VMA counter */
};


/**
 * @struct gl860_coord
 */
struct gl860_coord {
	int x;								/**< X-coordonate */
	int y;								/**< Y-coordonate */
};


/**
 * @struct gl860_video
 */
struct gl860_video {
	int fps;							/**< FPS setting */
	int brightness;						/**< Brightness setting */
	int contrast;						/**< Contrast setting */
	int whiteness;						/**< Whiteness setting */
	int colour;							/**< Colour setting */
	int depth;							/**< Depth colour setting */
	int palette;						/**< Palette setting */
	int hue;							/**< Hue setting */
	int hflip;							/**< Horizontal flip */
	int vflip;							/**< Vertical flip */
	int saturation;
	int srclum;
	int freq50;
	int net;
	int miroirh;
	int miroirv;
	int contrej;
	int v4l2;
	int retournee;
};


/**
 * @struct usb_gl860
 */
struct usb_gl860 {
	struct video_device *vdev; 			/**< Pointer on a V4L2 video device */
	struct usb_device *udev;			/**< Pointer on a USB device */
	struct usb_interface *interface;	/**< Pointer on a USB interface */

	int release;						/**< Release of the device (bcdDevice) */
	int webcam_model;					/**< Model of video camera device */
	int webcam_type;					/**< Type of camera : VGA, SXGA (1.3M), UXGA (2M) */
	unsigned char pilote;
	char doNotDisturb;
	char initFinie;
	int nbIm;
	int waitSet;

	unsigned char *int_in_buffer;		/**< Interrupt IN buffer */
	size_t int_in_size;					/**< Interrupt IN buffer size */
	__u8 int_in_endpointAddr;			/**< Interrupt IN endpoint address */

	size_t isoc_in_size;				/**< Isochrone IN size */
	__u8 isoc_in_endpointAddr;			/**< Isochrone IN endpoint address */

	int watchdog;						/**< Counter for the software watchdog */

	struct gl860_video vsettings;		/**< Video settings (brightness, whiteness...) */

	int error_status;

	int vopen;							/**< Video status (Opened or Closed) */
	int visoc_errors;					/**< Count the number of ISOCH errors */
	int vframes_error;					/**< Count the number of fault frames (so dropped) */
	int vframes_dumped;					/**< Count the number of ignored frames */


	spinlock_t spinlock;				/**< Spin lock */
	struct semaphore mutex;				/**< Mutex */
	wait_queue_head_t wait_frame;		/**< Queue head */


	// 1: isoc
	char isoc_init_ok;
	struct gl860_iso_buf isobuf[MAX_ISO_BUFS];
	int cam_off;
	int en_fct;
	int err45;
	int lealt;
	int offset;
	int nbPaq;

	// 2: frame
	int frame_size;
	int inter02;
	struct gl860_frame_buf *framebuf;
	struct gl860_frame_buf *empty_frames, *empty_frames_tail;
	struct gl860_frame_buf *full_frames, *full_frames_tail;
	struct gl860_frame_buf *fill_frame;
	struct gl860_frame_buf *read_frame;

	// 3: image
	int view_size;
	int image_size;
	void *image_data;
	struct gl860_image_buf images[WCAM_MAX_IMAGES];
	int image_used[WCAM_MAX_IMAGES];
	unsigned int nbuffers;
	unsigned int len_per_image;
	int image_read_pos;
	int fill_image;
	int resolution;
	struct gl860_coord view;
	struct gl860_coord image;
};


/**
 * @def GL860_PERCENT
 *   Calculate a value from a percent
 */
#define WCAM_PERCENT(x,y) ( ((int)x * (int)y) / 100)


/**
 * @def to_gl860_dev(d)
 * Cast a member of a structure out to the containing structure
 */
#define to_gl860_dev(d) container_of(d, struct usb_gl860, kref)


extern const struct gl860_coord gl860_image_sizes[WCAM_NBR_SIZES];

	
int fct_RTx(struct usb_gl860 *dev, char pref, unsigned int req,  __u16 val, __u16 index, int lg, void *pdonnees);
int fct_RTx2(struct usb_gl860 *dev, char pref, unsigned int req,  __u16 val, __u16 index, int lg, void *pdonnees, char *lgn);
int usb_gl860_write_registry(struct usb_gl860 *, __u16, __u16);
int usb_gl860_read_registry(struct usb_gl860 *, __u16, int *);
int usb_gl860_set_feature(struct usb_gl860 *, int);
int usb_gl860_set_configuration(struct usb_gl860 *);
int usb_gl860_isoc_init(struct usb_gl860 *);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
void usb_gl860_isoc_handler(struct urb *, struct pt_regs *);
#else
void usb_gl860_isoc_handler(struct urb *);
#endif
void usb_gl860_isoc_cleanup(struct usb_gl860 *);

int dev_gl860_devinette(struct usb_gl860 *dev, int vendor_id, int product_id, int id_pilote);
int dev_gl860_initialize_device(struct usb_gl860 *);
int dev_gl860_start_stream(struct usb_gl860 *);
int dev_gl860_stop_stream(struct usb_gl860 *);
int dev_gl860_check_device(struct usb_gl860 *, int);
int dev_gl860_camera_on(struct usb_gl860 *);
int dev_gl860_camera_off(struct usb_gl860 *);
int dev_gl860_camera_asleep(struct usb_gl860 *);
int dev_gl860_init_camera(struct usb_gl860 *);
int dev_gl860_reconf_camera(struct usb_gl860 *);
int dev_gl860_camera_settings(struct usb_gl860 *);
int dev_gl860_set_camera_quality(struct usb_gl860 *);
int dev_gl860_set_camera_fps(struct usb_gl860 *);
int dev_gl860_watchdog_camera(struct usb_gl860 *);

int v4l_gl860_select_video_mode(struct usb_gl860 *, int, int);
int v4l_gl860_register_video_device(struct usb_gl860 *);
int v4l_gl860_unregister_video_device(struct usb_gl860 *);

int gl860_create_sysfs_files(struct video_device *);
void gl860_remove_sysfs_files(struct video_device *);

int gl860_allocate_buffers(struct usb_gl860 *);
int gl860_reset_buffers(struct usb_gl860 *);
int gl860_clear_buffers(struct usb_gl860 *);
int gl860_free_buffers(struct usb_gl860 *);
void gl860_next_image(struct usb_gl860 *);
int gl860_next_frame(struct usb_gl860 *);
int gl860_handle_frame(struct usb_gl860 *);

int gl860_decompress(struct usb_gl860 *);


#endif 
